@@ -2,25 +2,33 @@
 
 use log::{error, warn};
 use std::cmp;
-use std::collections::BTreeMap;
+use std::collections::{BTreeMap, HashSet};
 use std::convert::TryInto;
+#[cfg(not(feature = "async"))]
 use std::fs::File;
+#[cfg(not(feature = "async"))]
 use std::io::Read;
 use std::path::Path;
 use std::sync::Mutex;
 
 #[cfg(feature = "rayon")]
 use rayon::prelude::*;
+#[cfg(feature = "async")]
+use tokio::fs::File;
+#[cfg(feature = "async")]
+use tokio::io::{AsyncRead, AsyncReadExt};
+#[cfg(feature = "async")]
+use tokio::pin;
 
-use super::parser;
-use super::{Document, Object, ObjectId};
 use crate::error::XrefError;
 use crate::object_stream::ObjectStream;
+use crate::parser;
 use crate::xref::XrefEntry;
-use crate::{Error, IncrementalDocument, Result};
+use crate::{Document, Error, IncrementalDocument, Object, ObjectId, Result};
 
 type FilterFunc = fn((u32, u16), &mut Object) -> Option<((u32, u16), Object)>;
 
+#[cfg(not(feature = "async"))]
 impl Document {
     /// Load a PDF document from a specified file path.
     #[inline]
@@ -62,6 +70,43 @@ impl Document {
     }
 }
 
+#[cfg(feature = "async")]
+impl Document {
+    pub async fn load<P: AsRef<Path>>(path: P) -> Result<Document> {
+        let file = File::open(path).await?;
+        let metadata = file.metadata().await?;
+        let capacity = Some(metadata.len() as usize);
+        Self::load_internal(file, capacity, None).await
+    }
+
+    pub async fn load_filtered<P: AsRef<Path>>(path: P, filter_func: FilterFunc) -> Result<Document> {
+        let file = File::open(path).await?;
+        let metadata = file.metadata().await?;
+        let capacity = Some(metadata.len() as usize);
+        Self::load_internal(file, capacity, Some(filter_func)).await
+    }
+
+    async fn load_internal<R: AsyncRead>(
+        source: R, capacity: Option<usize>, filter_func: Option<FilterFunc>,
+    ) -> Result<Document> {
+        pin!(source);
+
+        let mut buffer = capacity.map(Vec::with_capacity).unwrap_or_default();
+        source.read_to_end(&mut buffer).await?;
+
+        Reader {
+            buffer: &buffer,
+            document: Document::new(),
+        }
+        .read(filter_func)
+    }
+
+    /// Load a PDF document from a memory slice.
+    pub fn load_mem(buffer: &[u8]) -> Result<Document> {
+        buffer.try_into()
+    }
+}
+
 impl TryInto<Document> for &[u8] {
     type Error = Error;
 
@@ -74,6 +119,7 @@ impl TryInto<Document> for &[u8] {
     }
 }
 
+#[cfg(not(feature = "async"))]
 impl IncrementalDocument {
     /// Load a PDF document from a specified file path.
     #[inline]
@@ -108,6 +154,44 @@ impl IncrementalDocument {
     }
 }
 
+#[cfg(feature = "async")]
+impl IncrementalDocument {
+    /// Load a PDF document from a specified file path.
+    #[inline]
+    pub async fn load<P: AsRef<Path>>(path: P) -> Result<Self> {
+        let file = File::open(path).await?;
+        let metadata = file.metadata().await?;
+        let capacity = Some(metadata.len() as usize);
+        Self::load_internal(file, capacity).await
+    }
+
+    /// Load a PDF document from an arbitrary source.
+    #[inline]
+    pub async fn load_from<R: AsyncRead>(source: R) -> Result<Self> {
+        Self::load_internal(source, None).await
+    }
+
+    async fn load_internal<R: AsyncRead>(source: R, capacity: Option<usize>) -> Result<Self> {
+        pin!(source);
+
+        let mut buffer = capacity.map(Vec::with_capacity).unwrap_or_default();
+        source.read_to_end(&mut buffer).await?;
+
+        let document = Reader {
+            buffer: &buffer,
+            document: Document::new(),
+        }
+        .read(None)?;
+
+        Ok(IncrementalDocument::create_from(buffer, document))
+    }
+
+    /// Load a PDF document from a memory slice.
+    pub fn load_mem(buffer: &[u8]) -> Result<Document> {
+        buffer.try_into()
+    }
+}
+
 impl TryInto<IncrementalDocument> for &[u8] {
     type Error = Error;
 
@@ -146,13 +230,18 @@ impl<'a> Reader<'a> {
         let (mut xref, mut trailer) = parser::xref_and_trailer(&self.buffer[xref_start..], &self)?;
 
         // Read previous Xrefs of linearized or incremental updated document.
-        let mut prev_xref_start = trailer.remove(b"Prev");
-        while let Some(prev) = prev_xref_start.and_then(|offset| offset.as_i64().ok()) {
+        let mut already_seen = HashSet::new();
+        let mut prev_xref_start = trailer.get(b"Prev").cloned();
+        while let Ok(prev) = prev_xref_start.and_then(|offset| offset.as_i64()) {
+            if already_seen.contains(&prev) {
+                break;
+            }
+            already_seen.insert(prev);
             if prev < 0 || prev as usize > self.buffer.len() {
                 return Err(Error::Xref(XrefError::PrevStart));
             }
 
-            let (prev_xref, mut prev_trailer) = parser::xref_and_trailer(&self.buffer[prev as usize..], &self)?;
+            let (prev_xref, prev_trailer) = parser::xref_and_trailer(&self.buffer[prev as usize..], &self)?;
             xref.merge(prev_xref);
 
             // Read xref stream in hybrid-reference file
@@ -166,7 +255,7 @@ impl<'a> Reader<'a> {
                 xref.merge(prev_xref);
             }
 
-            prev_xref_start = prev_trailer.remove(b"Prev");
+            prev_xref_start = prev_trailer.get(b"Prev").cloned();
         }
 
         let xref_entry_count = xref.max_id() + 1;
@@ -189,7 +278,7 @@ impl<'a> Reader<'a> {
         let entries_filter_map = |(_, entry): (&_, &_)| {
             if let XrefEntry::Normal { offset, .. } = *entry {
                 let (object_id, mut object) = self
-                    .read_object(offset as usize, None)
+                    .read_object(offset as usize, None, &mut HashSet::new())
                     .map_err(|e| error!("Object load error: {:?}", e))
                     .ok()?;
                 if let Some(filter_func) = filter_func {
@@ -279,12 +368,22 @@ impl<'a> Reader<'a> {
         let object = self.document.get_object(object_id)?;
         let stream = object.as_stream()?;
 
-        stream.dict.get(b"Length").and_then(|value| {
-            if let Ok(id) = value.as_reference() {
-                return self.document.get_object(id).and_then(Object::as_i64);
-            }
-            value.as_i64()
-        })
+        stream
+            .dict
+            .get(b"Length")
+            .and_then(|value| {
+                if let Ok(id) = value.as_reference() {
+                    return self.document.get_object(id).and_then(Object::as_i64);
+                }
+                value.as_i64()
+            })
+            .map_err(|err| {
+                error!(
+                    "Stream dictionary of '{} {} R' is missing the Length entry",
+                    object_id.0, object_id.1
+                );
+                err
+            })
     }
 
     /// Get object offset by object id.
@@ -302,19 +401,26 @@ impl<'a> Reader<'a> {
         }
     }
 
-    pub fn get_object(&self, id: ObjectId) -> Result<Object> {
+    pub fn get_object(&self, id: ObjectId, already_seen: &mut HashSet<ObjectId>) -> Result<Object> {
+        if already_seen.contains(&id) {
+            warn!("reference cycle detected resolving object {} {}", id.0, id.1);
+            return Err(Error::ReferenceCycle);
+        }
+        already_seen.insert(id);
         let offset = self.get_offset(id)?;
-        let (_, obj) = self.read_object(offset as usize, Some(id))?;
+        let (_, obj) = self.read_object(offset as usize, Some(id), already_seen)?;
 
         Ok(obj)
     }
 
-    fn read_object(&self, offset: usize, expected_id: Option<ObjectId>) -> Result<(ObjectId, Object)> {
+    fn read_object(
+        &self, offset: usize, expected_id: Option<ObjectId>, already_seen: &mut HashSet<ObjectId>,
+    ) -> Result<(ObjectId, Object)> {
         if offset > self.buffer.len() {
             return Err(Error::Offset(offset));
         }
 
-        parser::indirect_object(self.buffer, offset, expected_id, self)
+        parser::indirect_object(self.buffer, offset, expected_id, self, already_seen)
     }
 
     fn get_xref_start(buffer: &[u8]) -> Result<usize> {
@@ -358,6 +464,7 @@ impl<'a> Reader<'a> {
     }
 }
 
+#[cfg(all(test, not(feature = "async")))]
 #[test]
 fn load_document() {
     let mut doc = Document::load("assets/example.pdf").unwrap();
@@ -369,6 +476,18 @@ fn load_document() {
     doc.save(file_path).unwrap();
 }
 
+#[cfg(all(test, feature = "async"))]
+#[tokio::test]
+async fn load_document() {
+    let mut doc = Document::load("assets/example.pdf").await.unwrap();
+    assert_eq!(doc.version, "1.5");
+
+    // Create temporary folder to store file.
+    let temp_dir = tempfile::tempdir().unwrap();
+    let file_path = temp_dir.path().join("test_2_load.pdf");
+    doc.save(file_path).unwrap();
+}
+
 #[test]
 #[should_panic(expected = "Xref(Start)")]
 fn load_short_document() {
